import os

from django.http import HttpResponse, HttpResponseBadRequest
from django.views import View


class DownloadFileView(View):

    def get(self, request):
        if 'download' not in request.session:
            return HttpResponseBadRequest()

        download = request.session.pop('download')

        with open(download['file_path'], 'rb') as fn:
            response = HttpResponse(fn.read())

        response['Content-Type'] = 'application/pdf'
        response['Content-Length'] = os.path.getsize(download['file_path'])
        response['Content-Disposition'] = 'attachment; filename="{}"'.format(download['file_name'])

        return response



""".. Ignore pydocstyle D400.

========
Exectuor
========

Executor is the central point which takes care to setup the runtime
environment and run all analyses.

At the moment only Docker executor is implemented:

.. autoclass:: lactolyse.executors.DockerExecutor
    :members:

"""
import atexit
import inspect
import logging
import os
import shutil
import tempfile
import time
from importlib import import_module

import docker

from lactolyse.analyses.base import BaseAnalysis

logger = logging.getLogger(__name__)

__all__ = ('executor',)

ANALYSES_PACKAGE = 'lactolyse.analyses'
DOCKER_IMAGE = 'latex:latest'
DOCKER_START_COMMAND = "/bin/sh -c 'sleep infinity'"
DOCKER_MOUNT_POINT = '/mnt'

docker_client = docker.from_env()


def remove_docker_container(container_id):

    def _remove_container():
        try:
            container = docker_client.containers.get(container_id)
        except docker.errors.NotFound:
            # Looks like container has already been removed.
            return

        # container.stop()
        container.remove(force=True)

    return _remove_container


class DockerExecutor():
    """Executor using Docker container for the environment.

    This executor starts a Docker container at the initiazlization and
    executes all commands inside it.
    """

    def __init__(self):
        self._runtime_root = tempfile.TemporaryDirectory()
        logger.info("Runtime root: {}".format(self._runtime_root.name))

        self._container = None
        self._analyses = {}

        self._create_container()
        self._discover_analyses()

    def _create_container(self):
        """Create the container."""
        runtime_mount = docker.types.Mount(DOCKER_MOUNT_POINT, self._runtime_root.name, 'bind')

        logger.debug("Creating container.")
        self._container = docker_client.containers.run(
            DOCKER_IMAGE, DOCKER_START_COMMAND, mounts=[runtime_mount], detach=True
        )
        logger.info("Container created: {}".format(self._container.id))

        # Remove the container on exit.
        atexit.register(remove_docker_container(self._container.id))

    def _validate_analysis(self, analysis):
        cls_name = analysis.__name__

        assert analysis.name, (
            "Subclass '{}' must have defined 'name' attribute.".format(cls_name)
        )
        assert analysis.template, (
            "Subclass '{}' must have defined 'template' attribute.".format(cls_name)
        )

    def _discover_analyses(self):
        """Iterate over files in analyses package and import all found analyses."""
        analyses_package = import_module(ANALYSES_PACKAGE)
        files = next(os.walk(analyses_package.__path__[0]))[2]

        for fn in files:
            fn = fn.rstrip('.py')
            module = import_module('{}.{}'.format(ANALYSES_PACKAGE, fn))

            for attr_name in dir(module):
                attr = getattr(module, attr_name)

                if (not inspect.isclass(attr) or
                    not issubclass(attr, BaseAnalysis) or
                    attr is BaseAnalysis):
                        continue

                self._validate_analysis(attr)
                self._analyses[attr.name] = attr

        logger.info(
            "Found {} analyses: {}".format(len(self._analyses), list(self._analyses.keys()))
        )

    def _check_container(self):
        """Check that container is up and running.

        Reload the container status and start it if it is not running
        already. If the container doesn't exist anymore, create a new
        one.
        """
        try:
            self._container.reload()
        except docker.errors.NotFound:
            logger.warning("Container cannot be found, creating a new one.")
            self._create_container()

            return

        if self._container.status != 'running':
            logger.warning("Container was not running, starting it.")
            self._container.start()

    def run(self, analysis, output_path, inputs=None):
        """Run the analysis."""
        if inputs is None:
            inputs = {}

        if not isinstance(inputs, dict):
            raise ValueError("Attribute 'inputs' must be of type 'dict' or 'None'.")

        if analysis not in self._analyses.keys():
            raise ValueError(
                "Unknown analysis '{}', select one of the following: {}".format(
                    analysis, ", ".join(self._analyses.keys())
                )
            )

        self._check_container()

        with tempfile.TemporaryDirectory(dir=self._runtime_root.name) as runtime_dir:
            analysis = self._analyses[analysis](runtime_dir=runtime_dir)

            start_time = time.time()

            logger.debug("Running run_render_context method.")
            analysis.run_render_context(inputs)

            logger.debug("Running render_template method.")
            analysis.render_template()

            logger.debug("Compiling report.")
            result = self._container.exec_run(
                "/bin/sh -c 'cd {}; {}'".format(
                    os.path.join(DOCKER_MOUNT_POINT, os.path.basename(runtime_dir)),
                    analysis.run_command,
                )
            )

            logger.info("Report finished in {:.2f}s.".format(time.time() - start_time))

            if result.exit_code != 0:
                message = "Something went wrong while rendering report."
                logger.error(message)
                print(result.output)
                raise RuntimeError(message)

            shutil.copy(analysis.get_pdf_file(), output_path)


executor = DockerExecutor()



from django.contrib import admin
from django.contrib.auth import views as auth_views
from django.urls import path
from django.views.generic import RedirectView

from . import views

urlpatterns = [
    path('login/', auth_views.login, {'template_name': 'lactolyse/login.html'}, name='login'),
    path('logout/', auth_views.logout, {'next_page': 'index'}, name='logout'),

    path('', RedirectView.as_view(pattern_name='select_analyses'), name="index"),
    path('tests/', views.select_analyses_view, name='select_analyses'),
    path('tests/lactate/', views.lactate_threshold_view, name='lactate_analyses'),
    path('tests/success/', views.analyses_success_view, name='analyses_success'),
    path('download/', views.download_file_view, name='download_file'),
]



import os

from django.contrib.auth.decorators import login_required
from django.views.generic import TemplateView

from .download import DownloadFileView
from .lactate_threshold import LactateThresholdView


select_analyses_view = login_required(
    TemplateView.as_view(template_name=os.path.join('lactolyse', 'select_analyses.html'))
)
analyses_success_view = login_required(
    TemplateView.as_view(template_name=os.path.join('lactolyse', 'analyses_success.html'))
)
download_file_view = DownloadFileView.as_view()
lactate_threshold_view = LactateThresholdView.as_view()



from django import forms

from material import Layout, Row

from .models import Athlete, LactateMeasurement


class AthleteForm(forms.ModelForm):

    class Meta:
        model = Athlete
        fields = ['name', 'age', 'weight']


class LactateMeasurementForm(forms.ModelForm):

    class Meta:
        model = LactateMeasurement
        fields = ['power', 'heart_rate', 'lactate']

    layout = Layout(
        Row('power', 'heart_rate', 'lactate')
    )

    def save(self, analyses, commit=True):

        self.instance.analyses = analyses

        return super().save(commit=commit)



import os
import tempfile

from django import forms
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.files import File
from django.db import transaction
from django.forms import formset_factory
from django.http import HttpResponse
from django.urls import reverse_lazy

from lactolyse.executors import executor
from lactolyse.forms import AthleteForm, LactateMeasurementForm
from lactolyse.models import LactateThresholdAnalyses

from .base import MultiFormView


class LactateThresholdView(LoginRequiredMixin, MultiFormView):
    template_name = os.path.join('lactolyse', 'lactate_threshold.html')
    analyses_name = 'lactate_threshold'
    success_url = reverse_lazy('analyses_success')

    form_classes = {
        'athlete': AthleteForm,
        'measurements': formset_factory(
            LactateMeasurementForm,
            min_num=5,
            extra=4,
            validate_min=True,
        ),
    }

    initial = {
        'athlete': {
            'name': "Domen Blenkuš",
            'age': "28",
            'weight': 79.5,
        },
        'measurements': [
            {
                'power': 80,
                'heart_rate': 125,
                'lactate': 2.4,
            }, {
                'power': 120,
                'heart_rate': 135,
                'lactate': 2.5,
            }, {
                'power': 160,
                'heart_rate': 147,
                'lactate': 1.6,
            }, {
                'power': 200,
                'heart_rate': 160,
                'lactate': 2.2,
            }, {
                'power': 240,
                'heart_rate': 177,
                'lactate': 4.1,
            }, {
                'power': 280,
                'heart_rate': 195,
                'lactate': 12.4,
            }
        ],
    }

    def _save_data(self, forms):
        with transaction.atomic():
            athlete = forms['athlete'].save()

            analyses = LactateThresholdAnalyses.objects.create(  # pylint: disable=no-member
                athlete=athlete
            )

            measurements = []
            for form in forms['measurements']:
                if form.cleaned_data:
                    measurement = form.save(analyses=analyses)
                    measurements.append(measurement)

        return analyses, athlete, measurements

    def _make_report(self, analyses, athlete, measurements):
        inputs = {
            'power': [int(m.power) for m in measurements],
            'heart_rate': [int(m.heart_rate) for m in measurements],
            'lactate': [float(m.lactate) for m in measurements],
            'weight': float(athlete.weight),
        }

        report_dir = tempfile.TemporaryDirectory()
        report_path = os.path.join(report_dir.name, 'report.pdf')

        executor.run(self.analyses_name, report_path, inputs)

        with open(report_path, 'rb') as fn:
            report_file = File(fn)
            analyses.report = report_file
            analyses.save()

    def forms_valid(self, forms):

        analyses, athlete, measurements = self._save_data(forms)
        self._make_report(analyses, athlete, measurements)

        # Set download file for download view.
        self.request.session['download'] = {
            'file_path': analyses.report.name,
            # 'file_name': "Report - {}.pdf".format(athlete.name),
            'file_name': "Report.pdf",
        }

        return super().forms_valid(forms)



import logging

import numpy as np

from .base import BaseAnalysis

logger = logging.getLogger(__name__)


class LactateThresholdAnalyses(BaseAnalysis):

    name = 'lactate_threshold'
    template = 'lactate_treshold.tex'

    def _calculate_dmax_context(self, inputs, lac_poly, hr_poly):

        def perpendicular(vector):
            return np.array(-vector[1], vector[0])

        # Determine where polynomial is the "most horizontal", i.e. root
        # of second derivative.

        min_x = np.roots(lac_poly.deriv())
        min_x = filter(lambda val: inputs['power'][0] < val < inputs['power'][-1], min_x)
        min_x = list(min_x)[0]

        # Find the point where polynomial starts to raise - threshold is
        # 0.5 - and take only real roots (hopefully there is only one).
        roots = np.roots(lac_poly - (lac_poly(min_x) + 0.5))
        roots = roots[np.isreal(roots)]
        roots = filter(lambda val: min_x < val < inputs['power'][-1], roots)
        roots = list(roots)

        # if len(roots) != 1:
        #     msg = "Something strange happened, there is more than one root."
        #     logger.error(msg)
        #     raise

        # Take the (real) root and calculate the y value.
        start_point = [roots[0].real, lac_poly(roots[0].real)]

        # Calculate the vector cross product.
        v_x = np.poly1d(inputs['power'][-1] - start_point[0])
        v_y = np.poly1d(lac_poly(inputs['power'][-1]) - start_point[1])
        u_x = np.poly1d([1, -start_point[0]])
        u_y = lac_poly - np.poly1d(start_point[1])
        cross_z = v_x * u_y - v_y * u_x

        ftp = np.roots(cross_z.deriv())
        ftp = filter(lambda val: start_point[0] < val < inputs['power'][-1], ftp)
        ftp = list(ftp)[0]

        line_poly = np.poly1d(np.polyfit(
            [start_point[0], inputs['power'][-1]],
            [start_point[1], inputs['lactate'][-1]],
            1
        ))

        perp_poly = perpendicular(line_poly)

        return {
            'power': ftp,
            'start_point': start_point,
            'end_point': [inputs['power'][-1], lac_poly(inputs['power'][-1])],
            'start_hr': hr_poly(start_point[0]),
            'heart_rate': hr_poly(ftp),
            'lactate': lac_poly(ftp),
        }

    def _calculate_cross_context(self, inputs, lac_poly, hr_poly):
        start_line = np.poly1d(
            np.polyfit(
                [inputs['power'][0], inputs['power'][0] + 5],
                [lac_poly(inputs['power'][0] ), lac_poly(inputs['power'][0] + 5)],
                1
            )
        )
        end_line = np.poly1d(
            np.polyfit(
                [inputs['power'][-1] - 5, inputs['power'][-1]],
                [lac_poly(inputs['power'][-1] - 5), lac_poly(inputs['power'][-1])],
                1
            )
        )

        cross = np.roots(start_line - end_line)
        power = cross[0]

        return {
            'power': power,
            'start_point': [inputs['power'][0], lac_poly(inputs['power'][0])],
            'end_point': [inputs['power'][-1], lac_poly(inputs['power'][-1])],
            'cross': [power, start_line(power)],
            'heart_rate': hr_poly(power),
            'lactate': lac_poly(power),
        }

    def _calculate_ftp_context(self, dmax, cross):
        return {
            'power': (dmax['power'] + cross['power']) / 2,
            'heart_rate': (dmax['heart_rate'] + cross['heart_rate']) / 2,
            'lactate': (dmax['lactate'] + cross['lactate']) / 2,
        }

    def _calculate_at_context(self, inputs, threshold, lac_poly, hr_poly):
        roots = np.roots(lac_poly - threshold)
        roots = roots[np.isreal(roots)]
        roots = filter(lambda val: inputs['power'][0] < val < inputs['power'][-1], roots)

        power = list(roots)[0].real

        return {
            'power': power,
            'heart_rate': hr_poly(power),
            'lactate': lac_poly(power),
        }

    def render_context(self, inputs):
        for attr in ['power', 'heart_rate', 'lactate']:
            if attr not in inputs:
                raise ValueError("Missing input '{}'.".format(attr))

        lac_poly = np.poly1d(np.polyfit(inputs['power'], inputs['lactate'], 3))
        hr_poly = np.poly1d(np.polyfit(inputs['power'], inputs['heart_rate'], 3))

        dmax_context = self._calculate_dmax_context(inputs, lac_poly, hr_poly)
        cross_context = self._calculate_cross_context(inputs, lac_poly, hr_poly)

        return {
            'inputs': inputs,
            'lac_poly': lac_poly,
            'dmax': dmax_context,
            'cross': cross_context,
            'ftp': self._calculate_ftp_context(dmax_context, cross_context),
            'at2': self._calculate_at_context(inputs, 2, lac_poly, hr_poly),
            'at4': self._calculate_at_context(inputs, 4, lac_poly, hr_poly),
        }



from django.apps import AppConfig


class LactolyseConfig(AppConfig):
    name = 'lactolyse'

    def ready(self):
        """Perform application initialization."""
        from .executors import executor



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Open source dataflow package for Django framework.
See:
https://github.com/genialis/resolwe
"""

from setuptools import find_packages, setup
# Use codecs' open for a consistent encoding
from os import path

base_dir = path.abspath(path.dirname(__file__))

# Get package metadata from 'lactolyse.__about__.py' file
about = {}
with open(path.join(base_dir, 'lactolyse', '__about__.py')) as f:
    exec(f.read(), about)

setup(
    name=about['__title__'],

    version=about['__version__'],

    description=about['__summary__'],

    url=about['__url__'],

    author=about['__author__'],
    author_email=about['__email__'],

    license=about['__license__'],

    # exclude tests from built/installed package
    packages=find_packages(exclude=['tests', 'tests.*', '*.tests', '*.tests.*']),
    install_requires=[
        'channels~=2.0.2',
        'Django~=2.0.0',
        'django-material~=1.2.2',
        'docker~=3.0.0',
        'Jinja2~=2.10',
        'numpy~=1.14.0',
        'psycopg2~=2.7.0',

    ],
    python_requires='>=3.6, <3.7',
    extras_require={
        'package': [
            'twine',
            'wheel',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',

        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Other Audience',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: OS Independent',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='ftp lactate threshold',
)



import os

import jinja2

TEX_JOB_NAME = 'report'

latex_jinja_env = jinja2.Environment(
    block_start_string='\BLOCK{',
    block_end_string='}',
    variable_start_string='\VAR{',
    variable_end_string='}',
    comment_start_string='\#{',
    comment_end_string='}',
    line_statement_prefix='%%',
    line_comment_prefix='%#',
    trim_blocks=True,
    autoescape=False,
    loader=jinja2.PackageLoader('lactolyse', os.path.join('templates', 'lactolyse', 'latex'))
)


class BaseAnalysis:

    def __init__(self, runtime_dir):
        self._runtime_dir = runtime_dir

        self._tex_file = os.path.join(self._runtime_dir, TEX_JOB_NAME + '.tex')
        self._pdf_file = os.path.join(self._runtime_dir, TEX_JOB_NAME + '.pdf')

        self._context = None
        self._rendered_template = None

    @property
    def name(self):
        """Name of the analysis."""
        raise NotImplementedError(
            "Subclass od 'BaseAnalysis' must define 'name' attribute."
        )

    @property
    def template(self):
        """Filename of the template file."""
        raise NotImplementedError(
            "Subclass od 'BaseAnalysis' must define 'template' attribute."
        )

    def get_pdf_file(self):
        """Return the path to the PDF file."""
        return self._pdf_file

    def render_context(self, inputs):
        """Calculate context for ``render_template`` method.

        This function should be overriden in subclass.
        """
        return {}

    def run_render_context(self, *args, **kwargs):
        """Run ``render_context`` and save the result.

        This is a utiliti function that simplify writing of
        ``render_context`` function and remove the need of directly
        accessing ``self._context`` argument.
        """
        self._context = self.render_context(*args, **kwargs)

    @property
    def run_command(self):
        return 'xelatex -interaction=nonstopmode -jobname={0} {0}.tex'.format(TEX_JOB_NAME)

    def render_template(self):
        template = latex_jinja_env.get_template(self.template)
        self._rendered_template = template.render(**self._context)

        with open(self._tex_file, 'w', encoding="utf-8") as fn:
            fn.write(self._rendered_template)
